//=========================================================== A2KArray
//
// Lesen und schreiben von DMNA-Dateien
//
// ===============================================================
//
// Copyright (C) Nds. Landesamt für Ökologie, Hildesheim, Germany, 2004
// Copyright (C) Landesumweltamt NRW, Essen, Germany, 2004
// Copyright (C) Landesanstalt für Umweltschutz, Karlsruhe, Germany, 2004
// Copyright (C) Janicke Consulting, 88662 Überlingen, Germany, 2004-2021
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//-------------------------------------------------------------------
//
// last change 2021-02-22 uj
//
//===================================================================

package de.janicke.tools;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

/**
 * The class A2KArray represents 1-, 2- or 3-dimensional arrays of floats. In
 * addition to the data a header is maintained that contains named parameters.
 * Parameter names are not case sensitive.
 * <p>
 * The array is either directly constructed or read in from a dmna/dmnb-file.
 * Within the header all lines starting with a blank or a hyphen are considered
 * as comments.
 */

public class A2KArray {

  /** Data values of the array (arranged in the same order as in C). */
  protected float[] data;

  /** Error message, if the constructor or a method failed. */
  public String message;

  /** Complete file name, if the array is read from a file. */
  public String name;

  /** Array type. null if not set in the file header. */
  public String artp;

  /**
   * Validity of the array. If the array can not be constructed properly
   * <code>valid</code> is set to <code>false</code> and the string
   * <code>message</code> contains an error message.
   */
  public boolean valid = false;

  /**
   * Header information.
   * 
   * The names (keys) are in lower case and the data objects are string arrays.
   */
  protected HashMap<String,String[]> values;

  /**
   * Type of definition within a grid cell. <br>
   * "P" = on grid points <br>
   * "V" = on cell midpoints (or as cell average) <br>
   */
  public String vldf;

  /** Overall factor */
  public float fact = 1;

  /** Format string */
  public String format;

  private static boolean DEBUG = false;
  private static boolean VERBOSE = false;

  // Number of array dimensions.
  private int dims;

  //  Lower and upper boundaries, number of index values.
  private int i1, i2, ni, j1, j2, nj, k1, k2, nk;

  //  Total number of elements.
  private int nn = -1;

  private int[] lowBound, hghBound;
  private int m0, mi, mj, mk;

  //================================================================ A2KArray
  /**
   * Read a A2K array from file. The array must be 1- or 2-dimensional with
   * elements of type <code>float</code> and in <code>text</code> mode. <br>
   * From the header the following parameters are evaluated: <br>
   * <code>dims, lowb, hghb, sequ, mode, vldf|valdef, xmin, ymin, delt|delta</code>
   * <br>
   * The values of <code>xmax</code> and <code>ymax</code> might be
   * specified as a second value after the value of <code>xmin</code> and of
   * <code>ymin</code>, otherwise they are calculated.
   * 
   * @param path
   *          name of the directory, where the file is located.
   * @param file
   *          name of the file (with extension).
   */
  public A2KArray(String path, String file) {
    if (VERBOSE) System.out.println("A2KArray(" + path + ", " + file + ")");
    File fl;
    FileReader flr;
    BufferedReader bfr;
    message = "starting";
    values = new HashMap<String,String[]>();
    fl = new File(path, file);
    name = fl.getPath();
    try {
      flr = new FileReader(fl);
      bfr = new BufferedReader(flr);
    } catch (IOException e) {
      message = "*** unable to read file " + name;
      if (VERBOSE) System.out.println("LasatArray: " + message);
      return;
    }

    String s, line;
    String[] ss = null, sqs = null;
    String sq = "";
    String check = "";
    int[] ii, iv = null;
    int nLines = 0;
    int i, ls, rc;
    boolean readingHeader;
    boolean bBinary = false;
    char c;
    float f;
    double[] dd;
    int n = 0;
    format = null;
    try {
      readingHeader = true;
      while (null != (line = bfr.readLine())) {
        nLines++;
        if (DEBUG) System.out.println("line " + nLines + ":" + line);
        if (line.length() < 1) continue;
        if (readingHeader) {
          c = line.charAt(0);
          if (c == '-') continue;
          if (c == ' ') continue;
          if (c == '*') {
            if (VERBOSE) System.out.println("VALUES:\n" + toString());
            check = "dims (number of dimensions)";
            ii = getIntegers("dims");
            if (ii == null) 
              break;
            dims = ii[0];
            if (dims < 1 || dims > 3) 
              break;
            check = "lowb (lower bounds)";
            ii = getIntegers("lowb");
            if (ii == null || ii.length != dims) 
              break;
            lowBound = ii;
            check = "hghb (upper bounds)";
            ii = getIntegers("hghb");
            if (ii == null || ii.length != dims) 
              break;
            hghBound = ii;
            check = "form";
            format = getString("form|format");
            LFormat lf = new LFormat(format);
            fact = (float)lf.factor;
            rc = setIndices();
            if (DEBUG) System.out.println("setIndices=" + rc);
            if (rc < 0) {
              System.out.println("*** error in array " + name + "\n"
                  + "can't set indices (" + rc + ")");
              valid = false;
              return;
            }
            check = "mode = text ?";
            s = getString("mode");
            bBinary = (s != null && !s.equals("text"));
            if (bBinary) {
              System.out.println("*** error in array " + name + "\n"
                  + "can't read binary data");
              valid = false;
              return;
            }
            check = "sequ (sequence definition)";
            sq = getString("sequ");
            if (sq == null) sq = "i+,j+,k+".substring(0, 3 * dims - 1);
            iv = new int[]{ i1, j1, k1 };
            sqs = getSqs(sq, iv);
            if (sqs == null) 
              break;
            check = "";
            artp = getString("arrtype|artp");
            if (artp != null) artp = artp.toUpperCase();
            s = getString("valdef|vldf");
            if (s != null)
              vldf = s;
            else {
              vldf = "P";
              if (artp != null && artp.startsWith("C")) vldf = "V";
            }
            dd = getDoubles("faktor|factor|fact");
            if (dd != null && dd.length > 0)
              fact = (float) dd[0];
            if (fact <= 0) fact = 1;
            c = vldf.charAt(0);
            readingHeader = false;
            data = new float[nn];
            if (DEBUG) System.out.println("header evaluated");
            continue;
          }
          check = "parse header";
          if (DEBUG) System.out.println("parsing header");
          if (!parseHeader(line, values)) break;
          if (DEBUG) System.out.println("continuing");
          continue;
        }
        check = "reading data";
        if (DEBUG) System.out.println("reading data");
        ss = line.split("\\s+");
        ls = ss.length;
        if (DEBUG) System.out.println("ls=" + ls);
        for (i = 0; i < ls; i++) {
          s = ss[i];
          if (s.length() == 0) continue;
          // if (DEBUG) System.out.println("s["+i+"]="+s+", iv="+iv);
          f = makeFloat(s);
          if (Float.isNaN(f)) break;
          if (iv == null) break;
          data[m0 + iv[0] * mi + iv[1] * mj + iv[2] * mk] = f/fact;
          iv = ivNext(iv, sqs);
          n++;
          if (n == nn) break;
        } // for i
        if (i < ls) break;
      } // while
      if (n == nn) valid = true;
      if (DEBUG) System.out.println("valid=" + valid);
    } // try
    catch (Exception e) {
      if (DEBUG) e.printStackTrace(System.out);
    }
    try {
      bfr.close();
      flr.close();
    } catch (IOException e) {
    }
    message = nLines + " lines read, nn=" + nn;
    if (!valid)
      message = "*** reading of " + name + " aborted with line " + nLines
          + "\n    error in check: " + check;
    else
      message = "reading of " + name + " succeeded (" + nLines + " lines)";
    if (VERBOSE) System.out.println(message);
  }

  private int[] ivNext(int[] iv, String[] sqs) {
    int ia = iv[0];
    int ja = iv[1];
    int ka = iv[2];
    int k;
    for (k = dims - 1; k >= 0; k--) {
      char c1 = sqs[k].charAt(0);
      char c2 = sqs[k].charAt(1);
      switch (c1) {
        case 'i' :
          if (c2 == '-') {
            ia--;
            if (ia < i1)
              ia = i2;
            else
              k = -1;
          } else {
            ia++;
            if (ia > i2)
              ia = i1;
            else
              k = -1;
          }
          break;
        case 'j' :
          if (c2 == '-') {
            ja--;
            if (ja < j1)
              ja = j2;
            else
              k = -1;
          } else {
            ja++;
            if (ja > j2)
              ja = j1;
            else
              k = -1;
          }
          break;
        case 'k' :
          if (c2 == '-') {
            ka--;
            if (ka < k1)
              ka = k2;
            else
              k = -1;
          } else {
            ka++;
            if (ka > k2)
              ka = k1;
            else
              k = -1;
          }
          break;
        default :
          k = 0;
          break;
      }
    } // for k
    if (k == -1) return null;
    return new int[]{ ia, ja, ka };
  }

  private String[] getSqs(String sq, int[] iv) {
    int ia, ja, ka, k;
    boolean[] indexSequ = new boolean[]{ false, false, false };
    ia = i1;
    ja = j1;
    ka = k1;
    String[] sqs = sq.split(",");
    if (sqs.length != dims) return null;
    for (k = 0; k < dims; k++) {
      if (sqs[k].length() < 2) sqs[k] += "+";
      char c1 = sqs[k].charAt(0);
      if (c1 == '+') return null;
      char c2 = sqs[k].charAt(1);
      switch (c1) {
        case 'i' :
          ia = (c2 == '-') ? i2 : i1;
          indexSequ[0] = true;
          break;
        case 'j' :
          ja = (c2 == '-') ? j2 : j1;
          indexSequ[1] = true;
          break;
        case 'k' :
          ka = (c2 == '-') ? k2 : k1;
          indexSequ[2] = true;
          break;
        default :
          return null;
      }
    }
    for (k = 0; k < dims; k++)
      if (!indexSequ[k]) return null;
    iv[0] = ia;
    iv[1] = ja;
    iv[2] = ka;
    return sqs;
  }

  /**
   * Create a 1-dimensional LASAT array. <br>
   * The parameters <code>i1</code> and <code>i2</code> are the lower and
   * upper boundaries of the array. Default settings are: <br>
   * <code>delt = 1</code>
   *  <code>xmin = 0</code>
   *  <code>xmax = i2-i1</code>
   *  <code>vldf = "P"</code>
   * 
   * @param i1
   *          lowest valid index value
   * @param i2
   *          highest valid index value
   */
  public A2KArray(int i1, int i2) {
    dims = 1;
    this.i1 = i1;
    this.i2 = i2;
    lowBound = new int[dims];
    hghBound = new int[dims];
    lowBound[0] = i1;
    hghBound[0] = i2;
    ni = i2 - i1 + 1;
    mi = 1;
    m0 = -(mi * i1);
    nn = ni;
    data = new float[nn];
    vldf = "P";
    valid = true;
    values = new HashMap<String,String[]>();
  }

  /**
   * Create a 2-dimensional LASAT array. <br>
   * The parameters <code>i1</code> and <code>i2</code> are the lower and
   * upper boundaries of the first array index, <code>j1</code> and
   * <code>j2</code> are the lower and upper boundaries of the second array
   * index. Default settings are: <br>
   * <code>delt = 1</code>
   *  <code>xmin = 0</code>
   *  <code>ymin = 0</code>
   *  <code>xmax = i2-i1</code>
   *  <code>ymax = j2-j1</code>
   *  <code>vldf = "P"</code>
   * 
   * @param i1
   *          lowest valid index value of first index
   * @param i2
   *          highest valid index value of first index
   * @param j1
   *          lowest valid index value of second index
   * @param j2
   *          highest valid index value of second index
   */
  public A2KArray(int i1, int i2, int j1, int j2) {
    dims = 2;
    this.i1 = i1;
    this.i2 = i2;
    this.j1 = j1;
    this.j2 = j2;
    lowBound = new int[dims];
    hghBound = new int[dims];
    lowBound[0] = i1;
    hghBound[0] = i2;
    lowBound[1] = j1;
    hghBound[1] = j2;
    ni = i2 - i1 + 1;
    nj = j2 - j1 + 1;
    mj = 1;
    mi = nj * mj;
    m0 = -(mi * i1 + mj * j1);
    nn = ni * nj;
    data = new float[nn];
    vldf = "P";
    valid = true;
    values = new HashMap<String,String[]>();
  }

  /**
   * Create a 3-dimensional LASAT array. <br>
   * The parameters <code>i1</code> and <code>i2</code> are the lower and
   * upper boundaries of the first array index, <code>j1</code> and
   * <code>j2</code> are the lower and upper boundaries of the second array
   * index, <code>k1</code> and <code>k2</code> are the lower and upper
   * boundaries of the third array index. Default settings are: <br>
   * <code>delt = 1</code>
   *  <code>xmin = 0</code>
   *  <code>ymin = 0</code>
   *  <code>xmax = i2-i1</code>
   *  <code>ymax = j2-j1</code>
   *  <code>vldf = "P"</code>
   * 
   * @param i1
   *          lowest valid index value of first index
   * @param i2
   *          highest valid index value of first index
   * @param j1
   *          lowest valid index value of second index
   * @param j2
   *          highest valid index value of second index
   * @param k1
   *          lowest valid index value of third index
   * @param k2
   *          highest valid index value of third index
   */
  public A2KArray(int i1, int i2, int j1, int j2, int k1, int k2) {
    dims = 3;
    this.i1 = i1;
    this.i2 = i2;
    this.j1 = j1;
    this.j2 = j2;
    this.k1 = k1;
    this.k2 = k2;
    lowBound = new int[dims];
    hghBound = new int[dims];
    lowBound[0] = i1;
    hghBound[0] = i2;
    lowBound[1] = j1;
    hghBound[1] = j2;
    lowBound[2] = k1;
    hghBound[2] = k2;
    ni = i2 - i1 + 1;
    nj = j2 - j1 + 1;
    nk = k2 - k1 + 1;
    mk = 1;
    mj = nk * mk;
    mi = nj * mj;
    m0 = -(mi * i1 + mj * j1 + mk * k1);
    nn = ni * nj * nk;
    data = new float[nn];
    vldf = "P";
    valid = true;
    values = new HashMap<String,String[]>();
  }

  /** Duplicates the array
   * @return 
   * @throws java.lang.CloneNotSupportedException */
  @Override
  public Object clone() throws CloneNotSupportedException {
    A2KArray la;
    if (dims == 1)
      la = new A2KArray(i1, i2);
    else if (dims == 2)
      la = new A2KArray(i1, i2, j1, j2);
    else if (dims == 3)
      la = new A2KArray(i1, i2, j1, j2, k1, k2);
    else
      return null;
    int n = data.length;
    la.data = new float[n];
    System.arraycopy(data, 0, la.data, 0, n);
    la.message = message;
    la.name = name;
    la.valid = valid;
    if (values != null) la.values = new HashMap<String,String[]>(values);
    la.vldf = vldf;
    la.format = format;
    la.dims = dims;
    la.i1 = i1;
    la.i2 = i2;
    la.ni = ni;
    la.j1 = j1;
    la.j2 = j2;
    la.nj = nj;
    la.k1 = k1;
    la.k2 = k2;
    la.nk = nk;
    la.nn = nn;
    la.lowBound = new int[dims];
    System.arraycopy(lowBound, 0, la.lowBound, 0, dims);
    la.hghBound = new int[dims];
    System.arraycopy(hghBound, 0, la.hghBound, 0, dims);
    la.m0 = m0;
    la.mi = mi;
    la.mj = mj;
    la.mk = mk;
    la.fact = fact;
    return la;
  }

  //--------------------------------------------------------- setIndices
  private int setIndices() {
    if (DEBUG) System.out.println("setting indices");
    if (dims < 1 || dims > 3) return -2;
    i1 = j1 = k1 = 0;
    i2 = j2 = k2 = 0;
    mi = mj = mk = 0;
    ni = nj = nk = 1;
    while (true) {
      i1 = lowBound[0];
      i2 = hghBound[0];
      ni = i2 - i1 + 1;
      mi = 1;
      if (dims == 1) break;
      j1 = lowBound[1];
      j2 = hghBound[1];
      nj = j2 - j1 + 1;
      mj = 1;
      mi *= nj;
      if (dims == 2) break;
      k1 = lowBound[2];
      k2 = hghBound[2];
      nk = k2 - k1 + 1;
      mk = 1;
      mi *= nk;
      mj *= nk;
      break;
    }
    nn = ni * nj * nk;
    m0 = -(mi * i1 + mj * j1 + mk * k1);
    return 0;
  }

  //----------------------------------------------------------- parseHeader
  private boolean parseHeader(String s, HashMap<String,String[]> hsh) {
    String[] ss;
    ArrayList<String> v = new ArrayList<String>();
    char c;
    String key;
    int i, j, l, n;
    l = s.length();
    for (i = 0; i < l; i++) {
      c = s.charAt(i);
      if (c == '\'') break;
      if (c == ' ' || c == '\t')
        continue;
      else {
        if (c == '\"') {
          for (j = i + 1; j < l; j++)
            if (s.charAt(j) == '\"' && (j == i + 1 || s.charAt(j - 1) != '\\'))
              break;
          v.add(s.substring(i, j + 1));
          j++;
        } else {
          for (j = i + 1; j < l; j++) {
            c = s.charAt(j);
            if (c == ' ' || c == '\t') break;
          }
          v.add(s.substring(i, j));
        }
      }
      i = j;
    }
    n = v.size();
    if (n == 0) return false;
    key = v.get(0);
    ss = new String[n - 1];
    for (i = 1; i < n; i++)
      ss[i - 1] = v.get(i);
    hsh.put(key.toLowerCase(), ss);
    return true;
  }

  //=============================================================== getDims
  /**
   * Get the number of dimensions of this array.
   * 
   * @return number of dimensions (currently only 1, 2 or 3).
   */
  public int getDims() {
    return dims;
  }

  //============================================================ getDoubles
  /**
   * Get the array of doubles assigned to the parameter <code>key</code>. The
   * Hashtable <code>values</code> is searched for the key <code>key</code>
   * and the array of strings assigned to this key is converted to doubles and
   * returned.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable is used.
   * @return the array of doubles or <code>null</code>, if the name is not
   *         found or a conversion to double is not possible.
   */
  public double[] getDoubles(String key) {
    int i, n;
    String[] ss = getStrings(key);
    if (ss == null) return null;
    n = ss.length;
    double[] dd = new double[n];
    for (i = 0; i < n; i++)
      dd[i] = makeDouble(ss[i]);
    return dd;
  }

  //============================================================ getDouble
  /**
   * Get the array of doubles assigned to the parameter <code>key</code>. The
   * Hashtable <code>values</code> is searched for the key <code>key</code>
   * and the string assigned to this key is converted to double and returned.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable ist used.
   * @return the double value or <code>NaN</code>, if the name is not found
   *         or a conversion to double is not possible.
   */
  public double getDouble(String key) {
    String s = getString(key);
    if (s == null) return Double.NaN;
    return makeDouble(s);
  }

  //=========================================================== getIntegers
  /**
   * Get the array of integers assigned to the parameter <code>key</code>.
   * The Hashtable <code>values</code> is searched for the key
   * <code>key</code> and the array of strings assigned to this key is
   * converted to integers and returned.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable ist used.
   * @return the array of integers or <code>null</code>, if the name is not
   *         found or a conversion to integer is not possible.
   */
  public int[] getIntegers(String key) {
    int i, n;
    String[] ss = getStrings(key);
    if (ss == null) return null;
    n = ss.length;
    int[] ii = new int[n];
    try {
      for (i = 0; i < n; i++)
        ii[i] = Integer.parseInt(ss[i].trim());
    } catch (NumberFormatException e) {
      return null;
    }
    return ii;
  }

  //=========================================================== getInteger
  /**
   * Get the integer assigned to the parameter <code>key</code>. The
   * Hashtable <code>values</code> is searched for the key <code>key</code>
   * and the string assigned to this key is converted to integer and returned.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable ist used.
   * @return the integer or <code>MIN_VALUE</code>, if the name is not found
   *         or a conversion to integer is not possible.
   */
  public int getInteger(String key) {
    int i;
    String s = getString(key);
    if (s == null) return Integer.MIN_VALUE;
    try {
      i = Integer.parseInt(s.trim());
    } catch (NumberFormatException e) {
      i = Integer.MIN_VALUE;
    }
    return i;
  }

  //======================================================= getLowerBound
  /**
   * Get the lowest index values of this array.
   * 
   * @return lowest valid index values
   */
  public int[] getLowerBounds() {
    return lowBound;
  }

  /**
   * Get the lowest index value of this array.
   * 
   * @param dim
   *          dimension count (starting with 0)
   * @return lowest valid index value for dimension <code>dim</code>
   * @throws java.lang.Exception
   */
  public int getLowerBound(int dim) throws Exception {
    int n = 0;
    if (dim < 0 || dim >= dims) exit("invalid dimension");
    switch (dim) {
      case 0 :
        n = i1;
        break;
      case 1 :
        n = j1;
        break;
    }
    return n;
  }

  public int getOffset(String s) {
    if (s.equals("0"))
      return m0;
    else if (s.equals("i"))
      return mi;
    else if (s.equals("j"))
      return mj;
    else if (s.equals("k")) return mk;
    return 0;
  }

  //=================================================== getNumberOfElements
  /**
   * Get the total number of elements.
   * 
   * @return total number of elements in this array.
   */
  public int getNumberOfElements() {
    return nn;
  }

  //============================================================ getStrings
  /**
   * Get the tokens assigned to the parameter <code>key</code>. The Hashtable
   * <code>values</code> is searched for the key <code>key</code> and the
   * array of strings assigned to this key is returned. Quotation marks are
   * removed.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable ist used.
   * @return the array of strings or <code>null</code>, if the name is not
   *         found.
   */
  public String[] getStrings(String key) {
    int kk1, kk2, l;
    l = key.length();
    for (kk1 = 0; kk1 < l; kk1++) {
      if (key.charAt(kk1) == '|') continue;
      kk2 = key.indexOf('|', kk1);
      if (kk2 < 0) kk2 = l;
      String[] ss = (String[]) values.get(key.substring(kk1, kk2).toLowerCase());
      if (ss != null) {
        for (int i = 0; i < ss.length; i++) {
          int ii2 = ss[i].length() - 1;
          if (ii2 >= 0 && ss[i].charAt(0) == '\"') { // unquote
            if (ss[i].charAt(ii2) == '\"') ss[i] = ss[i].substring(1, ii2);
          }
        }
        return ss;
      }
      kk1 = kk2;
    }
    return null;
  }

  //============================================================ getString
  /**
   * Get the first token assigned to the parameter <code>key</code>. The
   * Hashtable <code>values</code> is searched for the key <code>key</code>
   * and the first element of the array of strings assigned to this key is
   * returned. Quotation marks are removed.
   * 
   * @param key
   *          parameter name (not case sensitive). Several names may be
   *          specified, separated by the vertical bar "|". The first name found
   *          in the Hashtable ist used.
   * @return the string (without quotation marks) or <code>null</code>, if
   *         the name is not found or no value is defined.
   */
  public String getString(String key) {
    int kk1, kk2, l;
    String s = null;
    l = key.length();
    for (kk1 = 0; kk1 < l; kk1++) {
      if (key.charAt(kk1) == '|') continue;
      kk2 = key.indexOf('|', kk1);
      if (kk2 < 0) kk2 = l;
      String[] ss = (String[]) values.get(key.substring(kk1, kk2).toLowerCase());
      if (ss != null && ss.length != 0) {
        s = ss[0];
        int ii2 = s.length() - 1;
        if (ii2 >= 0 && s.charAt(0) == '\"') { // unquote
          if (s.charAt(ii2) == '\"') s = s.substring(1, ii2);
        }
        return s;
      }
      kk1 = kk2;
    }
    return null;
  }

  //======================================================= getUpperBound
  /**
   * Get the highest index values of this array.
   * 
   * @return highest valid index values
   */
  public int[] getUpperBounds() {
    return hghBound;
  }

  /**
   * Get the highest index value of this array.
   * 
   * @param dim
   *          dimension count (starting with 0)
   * @return highest valid index value for dimension <code>dim</code>
   * @throws java.lang.Exception
   */
  public int getUpperBound(int dim) throws Exception {
    int n = 0;
    if (dim < 0 || dim >= dims) exit("invalid dimension");
    switch (dim) {
      case 0 :
        n = i2;
        break;
      case 1 :
        n = j2;
        break;
    }
    return n;
  }

  /**
   * Get the index offset
   * 
   * @param flag
   *          return mj (flag > 0) or m0 (otherwise)
   * @return mj or m0
   */
  public int getIndexOffset(int flag) {
    if (flag > 0)
      return mj;
    else
      return m0;
  }

  //=========================================================== makeDouble
  /**
   * Convert the string <code>s</code> to a number of type <code>double</code>.
   * The following special cases are recognized: <br>
   * <code>undef</code> is converted to POSITIVE_INFINITY <br>
   * <code> +inf</code> is converted to POSITIVE_INFINITY <br>
   * <code> -inf</code> is converted to NEGATIVE_INFINITY <br>
   * <code>   ? </code> is converted to NEGATIVE_INFINITY <br>
   * 
   * @param s
   *          string to be converted
   * @return the converted number or NaN if no conversion is possible
   */
  static public double makeDouble(String s) {
    double d = Double.NaN;
    if (s != null) {
      s = s.toLowerCase();
      if (s.equals("undef"))
        d = Double.POSITIVE_INFINITY;
      else if (s.equals("+inf"))
        d = Double.POSITIVE_INFINITY;
      else if (s.equals("-inf"))
        d = Double.NEGATIVE_INFINITY;
      else if (s.equals("?"))
        d = Double.NEGATIVE_INFINITY;
      else {
        try {
          d = Double.parseDouble(s.replace(',', '.'));
        } catch (NumberFormatException e) {
        }
      }
    }
    return d;
  }

  //=========================================================== makeFloat
  /**
   * Convert the string <code>s</code> to a number of type <code>double</code>.
   * The following special cases are recognized: <br>
   * <code>undef</code> is converted to POSITIVE_INFINITY <br>
   * <code> +inf</code> is converted to POSITIVE_INFINITY <br>
   * <code> -inf</code> is converted to NEGATIVE_INFINITY <br>
   * <code>   ? </code> is converted to NEGATIVE_INFINITY <br>
   * 
   * @param s
   *          string to be converted
   * @return the converted number or NaN if no conversion is possible
   */
  static public float makeFloat(String s) {
    float d = Float.NaN;
    if (s != null) {
      s = s.toLowerCase();
      if (s.equals("undef"))
        d = Float.POSITIVE_INFINITY;
      else if (s.equals("+inf"))
        d = Float.POSITIVE_INFINITY;
      else if (s.equals("-inf"))
        d = Float.NEGATIVE_INFINITY;
      else if (s.equals("?"))
        d = Float.NEGATIVE_INFINITY;
      else {
        try {
          d = Float.parseFloat(s.replace(',', '.'));
        } catch (NumberFormatException e) {
        }
      }
    }
    return d;
  }

  //============================================================= setValue

  /**
   * Set the value at index <code>i</code>. If the array is not 1-dimensional
   * or the index value is invalid, the program is exited and a backtrace is
   * generated.
   * @param i
   * @param f
   * @throws java.lang.Exception
   */
  public void setValue(int i, float f) throws Exception {
    if (dims > 1) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    data[m0 + i * mi] = f;
  }

  /**
   * Set the value at index <code>i</code> and <code>j</code>. If the array
   * is not 2-dimensional or one of the index values is invalid, the program is
   * exited and a backtrace is generated.
   * @param i
   * @param j
   * @param f
   * @throws java.lang.Exception
   */
  public void setValue(int i, int j, float f) throws Exception {
    if (dims > 2) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    if (j < j1 || j > j2) exit("index j out of range");
    data[m0 + i * mi + j * mj] = f;
  }

  /**
   * Set the value at index <code>i</code> and <code>j</code>. If the array
   * is not 2-dimensional or one of the index values is invalid, the program is
   * exited and a backtrace is generated.
   * @param i
   * @param j
   * @param k
   * @param f
   * @throws java.lang.Exception
   */
  public void setValue(int i, int j, int k, float f) throws Exception {
    // if (dims != 2) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    if (j < j1 || j > j2) exit("index j out of range");
    if (k < k1 || k > k2) exit("index k out of range");
    data[m0 + i * mi + j * mj + k * mk] = f;
  }

  //================================================================ value
  /**
   * Get the value at index <code>i</code>. If the array is not 1-dimensional
   * or the index value is invalid, an exception is generated.
   * @param i
   * @return 
   * @throws java.lang.Exception
   */
  public float value(int i) throws Exception {
    if (dims > 1) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    return data[m0 + i * mi];
  }

  /**
   * Get the value at index <code>i</code> and <code>j</code>. If the array
   * is not 2-dimensional or one of the index values is invalid, an exception is
   * generated.
   * @param i
   * @param j
   * @return 
   * @throws java.lang.Exception
   */
  public float value(int i, int j) throws Exception {
    if (dims > 2) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    if (j < j1 || j > j2) exit("index j out of range");
    return data[m0 + i * mi + j * mj];
  }

  /**
   * Get the value at index <code>i</code> and <code>j</code>. If the array
   * is not 3-dimensional or one of the index values is invalid, an exception is
   * generated.
   * @param i
   * @param j
   * @param k
   * @return 
   * @throws java.lang.Exception 
   */
  public float value(int i, int j, int k) throws Exception {
    // if (dims != 3) exit("invalid dimension");
    if (i < i1 || i > i2) exit("index i out of range");
    if (j < j1 || j > j2) exit("index j out of range");
    if (k < k1 || k > k2) exit("index k out of range");
    return data[m0 + i * mi + j * mj + k * mk];
  }

  //---------------------------------------------------------------- exit
  private void exit(String s) throws Exception {
    throw new Exception("*** error in array " + name + "\n" + s);
  }

  //============================================================= toString
  /**
   * Converts the header into a string. The string consists of lines separated
   * by "\n". Each line contains the parameter name and the parameter values
   * separated by tabs.
   * 
   * @return string representation of the header.
   */
  @Override
  public String toString() {
    String[] ss;
    int i;
    StringBuilder b = new StringBuilder();
    Set<String> enm = values.keySet();
    String key = "";
    try {
      for (String s : enm) {
        key = s;
        ss = values.get(key);
        b.append(key).append("\t");
        for (i = 0; i < ss.length; i++)
          b.append(ss[i]).append("\t");
        b.append("\n");
      }
    } catch (Exception e) {
      System.out.println("key=" + key);
      e.printStackTrace(System.out);
    }
    return b.toString();
  }

  //-------------------------------------------------------- getMinimum
  public float getMinimum() throws Exception {
    int i, j;
    int ii1 = getLowerBound(0);
    int jj1 = getLowerBound(1);
    int ii2 = getUpperBound(0);
    int jj2 = getUpperBound(1);
    float min = Float.MAX_VALUE;
    float d;
    for (i = ii1; i <= ii2; i++) {
      for (j = jj1; j <= jj2; j++) {
        d = value(i, j);
        if (d < min) min = d;
      }
    }
    return min;
  }

  //-------------------------------------------------------- getMaximum
  public float getMaximum() throws Exception {
    int i, j;
    int ii1 = getLowerBound(0);
    int jj1 = getLowerBound(1);
    int ii2 = getUpperBound(0);
    int jj2 = getUpperBound(1);
    float max = -Float.MAX_VALUE;
    float d;
    for (i = ii1; i <= ii2; i++) {
      for (j = jj1; j <= jj2; j++) {
        d = value(i, j);
        if (d > max) max = d;
      }
    }
    return max;
  }

  //---------------------------------------------------------- getTable
  public HashMap<String,String[]> getTable() {
    return values;
  }

  //---------------------------------------------------------- setTable
  public void setTable(Map<String,String[]> m) {
    if (values == null) values = new HashMap<String,String[]>();
    values.putAll(m);
  }

  public void put(String key, String value) {
    values.put(key, new String[]{ value });
  }

  public void put(String key, String[] value) {
    values.put(key, value);
  }

  //------------------------------------------------------------ save
  public void save(String file) throws Exception {
    if (DEBUG) System.out.println("writing " + file);
    if (values == null) values = new HashMap<String,String[]>();
    StringBuilder sb = new StringBuilder();
    String s = null;
    put("dims", "" + dims);
    for (int i = 0; i < dims; i++) {
      if (i > 0) sb.append('\t');
      sb.append(lowBound[i]);
    }
    put("lowb", sb.toString().split("\t"));
    sb.setLength(0);
    for (int i = 0; i < dims; i++) {
      if (i > 0) sb.append('\t');
      sb.append(hghBound[i]);
    }
    put("hghb", sb.toString().split("\t"));
    sb.setLength(0);
    put("mode", "\"text\"");
    String sequ = getString("sequ");
    if (dims != 2) sequ = null;
    if (sequ != null && !sequ.equals("i+,j+")) sequ = null;
    String axes = getString("axes");
    if (sequ == null) {
      if (dims == 3) {
        put("sequ", "\"k+,j-,i+\"");
        if (axes != null) put("axes", "\"xyz\"");
      } else {
        if (axes != null) put("axes", "\"xy\"");
        put("sequ", "\"j-,i+\"");
      }
    }
    if (!file.endsWith(".dmna")) file += ".dmna";
    String form = getString("form|format");
    if (form == null) 
      throw new Exception("*** no data format");
    LFormat lf = new LFormat(form);
    fact = (float)lf.factor;
    String of = "\t" + lf.pformat;
    String locl = getString("locl");
    //boolean german = (locl == null || locl.equals("german"));
    boolean german = false;                                       //-2012-09-04
    if (locl != null && (locl.equalsIgnoreCase("german") || locl.equalsIgnoreCase("de")))
      german = true;
    BufferedWriter bw = null;
    bw = new BufferedWriter(new FileWriter(file));
    bw.write(toString());
    bw.write("*\n");
    if (sequ == null) {
      for (int k = k1; k <= k2; k++) {
        for (int j = j2; j >= j1; j--) {
          for (int i = i1; i <= i2; i++) {
            s = String.format(Locale.UK, of, fact*data[m0 + i*mi + j*mj + k*mk]);
            if (german)  s = s.replace('.', ',');
            bw.write(s);
          }
          bw.write("\n");
        }
        bw.write("\n");
      }
    } 
    else {
      for (int i = i1; i <= i2; i++) {
        for (int j = j1; j <= j2; j++) {
          s = String.format(Locale.UK, of, fact*data[m0 + i*mi + j*mj]);
          if (german)  s = s.replace('.', ',');
          bw.write(s);
        }
        bw.write("\n");
      }
    }
    bw.write("***\n");
    bw.close();
  }

  //=====================================================================
  /**
   * Test program. The array <code>s[1]</code> is read from the directory
   * <code>s[0]</code> and printed (sequence: i+,j+,k+).
   * @param s
   */
  public static void main(String s[]) {
    System.out.println("START");
    DEBUG = true;
    VERBOSE = true;
    if (s.length < 2) {
      System.out.println("Keine Aufrufparameter angegeben");
      System.exit(0);
    }
    try {
      A2KArray ar = new A2KArray(s[0], s[1]);
      System.out.println("valid=" + ar.valid);
      System.out.println(ar.toString());
      if (ar.valid) {
        System.out.println("i1=" + ar.i1 + ", i2=" + ar.i2);
        System.out.println("j1=" + ar.j1 + ", j2=" + ar.j2);
        System.out.println("k1=" + ar.k1 + ", k2=" + ar.k2);
        for (int i = ar.i1; i <= ar.i2; i++) {
          for (int j = ar.j1; j <= ar.j2; j++) {
            for (int k = ar.k1; k <= ar.k2; k++) {
              float f = ar.value(i, j, k);
              System.out.println("(" + i + "," + j + "," + k + ") " + f);
            }
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace(System.out);
    }
    System.out.println("finished");
    System.exit(0);
  }
  
  //================================================================= LFormat
  
  private static class LFormat {
    
    String form;
    String pformat;
    String name;
    double factor=1;
    
    public LFormat(String form) throws Exception {
      this.form = form;
      int i1 = form.indexOf('%');
      name = (i1 > 0) ? form.substring(0, i1) : "";
      if (form.charAt(i1 + 1) == '(') {
        if (form.charAt(i1+2) != '*')
          throw new Exception("invalid format factor!");
        int i2 = form.indexOf(')');
        String t = form.substring(i1+3, i2);
        factor = Double.parseDouble(t.replace(',', '.'));
        if (form.charAt(i1+1) == '/')  factor = 1/factor;
        pformat = "%" + form.substring(i2+1);
      }
      else {
        pformat = form.substring(i1);
      }
    }
    
  }

} ////////////////////////////////////////////////////////////////////////

